//=============================================================================
// Given a pointer to text, these routines tokenize the data for evaluation
//
// Written by Lane Roathe
// (c) Copyright 1995 Ideas From the Deep, Ltd. All Rights Reserved!
//
// RELEASE NOTES:
//
//	This code is released as CodeWare, meaning that you are free to use it,
//	in exchange for sending Lane Roathe <lroathe@ifd.fixnet.com> a utility and/or
//	piece of source code of similar value. You may distribute this code freely,
//	as long as it is distributed as an entire package without modifications!
//
//	If you use any of this code, please give Lane Roathe credit. Thanks!
//
//	If you make modifications you'd like to see in the distribution package,
//	please feel free to send them to Lane w/a note. Most likely it will happen.
//---------------------------------------------------------------------
// 03-18-97 LR :fix '}' at end of expression (ie, in an ENUM)
// 12-20-95 LR :remove ctype.c stuff, window include file problems!
// 06-12-95 LR :creation for DASM, seems to work...
//=============================================================================

#include "includes.h"	// required to create PCH on PC, empty on Mac

#define sbump (*s)++

//=============================================================================
//	TokenizeExpression
//
//	Creates a token stream from text for evaluation
//=============================================================================

OSErr TokenizeExpression( token * * t , const char * * s )
{
	OSErr error;
	int	numParens = 0;	/* parenthesis balance */
	int eol = false, done = false;

	while ( !eol && !done )			/* stream through bytes */
	{
		gpErrWord = (Ptr)*s;

		switch (*sbump)
		{
			case ' ':		/* skip any whitespace */
			case '\t':
				break;

			case '\'':
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
			case '$':
			case '%':
				(*s)--;	// put back that character
				AddToken(t,TOKEN_CONST);
				if( (error = AddConst(t,s)) != noErr )
					return( error );
				break;

			case '=':
				if(**s == '=')
				 {
					sbump;
					AddToken(t,TOKEN_EQUAL);
				 }
				else
					return ERROR_MISSINGPARAM;
				break;

			case '!':
				if(**s == '=')
				 {
					sbump;
					AddToken(t,TOKEN_NOTEQUAL);
				 }
				else
					return ERROR_MISSINGPARAM;
//				AddToken(t,TOKEN_LOGICALNOT);	//LR not yet implemented in Evaluate
				break;

			case '<':
				if(**s == '<')
				 {
					sbump;
					AddToken(t,TOKEN_LSHIFT);
				 }
				else if(**s == '=')
				 {
					sbump;
					AddToken(t,TOKEN_LESSTHANOREQUAL);
				 }
				else
					AddToken(t,TOKEN_LESSTHAN);
				break;

			case '>':
				if(**s == '>')
				 {
					sbump;
					AddToken(t,TOKEN_RSHIFT);
				 }
				else if(**s == '=')
				 {
					sbump;
					AddToken(t,TOKEN_GREATERTHANOREQUAL);
				 }
				else
					AddToken(t,TOKEN_GREATERTHAN);
				break;

			case '+':
				AddToken(t,TOKEN_PLUS);
				break;

			case '-':
				AddToken(t,TOKEN_MINUS);
				break;

			case '*':
				AddToken(t,TOKEN_MULT);
				break;

			case '/':
				if( **s == '/' )
					eol = true;
				else
					AddToken(t,TOKEN_DIV);
				break;

			case '(':
				AddToken(t,TOKEN_LPAREN);
				numParens++;
				break;

			case ')':
				AddToken(t,TOKEN_RPAREN);
				numParens--;
				break;

			case '&':
				AddToken(t,TOKEN_AND);
				break;

			case '|':
				AddToken(t,TOKEN_OR);
				break;

			case '^':
				AddToken(t,TOKEN_CARET);
				break;

			case '\\':
				AddToken(t,TOKEN_MOD);
				break;

			case '~':
				AddToken(t,TOKEN_NOT);
				break;

			case 0:		// null
			case 13:	// carriage return
			case 10:	// line feed
			case ';':	// ending comment
				eol = true;
				break;

			case ',':		// eat comma, but leave ending braces!
				(*s)++;
			case '}':
				done = true;	// done, but not EOL (and must bypass comma!)
				break;

			default:					// assume a variable
				(*s)--;					// put 1st char of vname back
				AddToken(t,TOKEN_VAR);
				if( (error = AddVar(t,s)) != noErr )
					return( error );
				break;
			}
		}

	(*s)--;
	AddToken(t,TOKEN_END);	// backup one char for NextLine call & signal no more tokens

	if (numParens)
		return ERROR_MISBALPAREN;	// misbalanced parens

	if( eol )
		return( ERROR_EOL );
	else
		return( NOERR );
}

//=============================================================================
//	AddToken
//
//	Adds a token to the token buffer
//=============================================================================

void AddToken(token * * stream, token t)	/* add a token to the stream */
{
		*(*stream)++ = t;
}

//=============================================================================
//	AddConst
//
//	Adds a constant to the token buffer
//=============================================================================

OSErr AddConst(token * * stream, const char * * s )
{
	unsigned long	i;
	OSErr error = NOERR;

	switch( **s )
	{
		case '$':
			(*s)++;
			i = HexLong(s);
			break;
		case '%':
			(*s)++;
			i = BinLong(s);
			break;
		case '\'':
			(*s)++;
			i = ASCIILong(s,&error);
			break;
		default:
			if( *((*s) + 1) == 'x' ||  *((*s) + 1) == 'X' )
			{
				(*s) += 2;
				i = HexLong( s );
			}
			else
				i = DecLong( s );
			break;
	}

	if( i <= MAXBYTE )
	{
		AddToken ( stream, CLASS_BYTE );
		AddToken ( stream, i );
	}
	else if( i <= MAXWORD )
	{
		AddToken (stream, CLASS_WORD );
		AddWord (stream, i);
	}
	else
	{
		AddToken(stream, CLASS_LONG);
		AddLong (stream, i);
	}

	return(error);
}

//=============================================================================
//	AddWord
//
//	Given a word, adds it the token buffer as two bytes (lo/hi always!)
//=============================================================================

void AddWord ( token * * stream, long i )
{
	AddToken (stream, i & MAXBYTE );
	AddToken (stream, i >> 8 );
}

//=============================================================================
//	AddLong
//
//	Given a long, adds it the token buffer as four bytes (lo..hi always!)
//=============================================================================

void AddLong ( token * * stream, long i )
{
	int	j;

	for ( j = 0 ; j < 4 ; j++ ) {
		AddToken ( stream, i & MAXBYTE );
		i >>= 8 ;
		}
}

//=============================================================================
//	BinLong
//
//	Create a long binary from char data
//=============================================================================

long BinLong ( const char * * s )
{
	register long r = 0;

	while ( **s == '1' || **s == '0' )
	{
		r <<= 1;
		r += (**s) - '0';
		(*s)++;
	}

	return r;
}

//=============================================================================
//	ASCIILong
//
//	Given a 'xxxx' char stream, create a long value (ie, ResEdit type)
//=============================================================================

long ASCIILong ( const char * * s , OSErr *error)
{
	register int c = 0;
	register long r = 0;

	while( (**s) && ((**s) != '\'') && c < 4 )
	{
		r <<= 8;
		r |= (**s);
		(*s)++;
		c++;
	}

	if( **s && c <= 4 )
		(*s)++;
	else
		*error = ERROR_SYNTAX;
	return r;
}

//=============================================================================
//	DecLong
//
//	Create a long decimal from ascii text
//=============================================================================

long DecLong ( const char * * s )
{
	long r = 0;

	while ( **s >= '0' && **s <= '9' )
	{
		r *= 10L;
		r += (**s) - '0';
		(*s)++;
	}

	return r;
}

//=============================================================================
//	HexLong
//
//	Create a decimal long from ascii text specifying Hexidecimal values
//=============================================================================

static Boolean ishex( char c )
{
	return( ( c >= '0' && c <= '9' ) || ( c >= 'A' && c <= 'F' ) );
}

static Byte hexdigit( char c )
{
	if( c >= '0' && c <= '9' )	// 0 - 9
		return( c - '0' );

	if( c >= 'A' && c <= 'F' )	// 10-15
		return( c - 'A' + 10 );

	return( 0 );
}

long HexLong ( const char * * s )
{
	long r = 0;

	while( ishex( upper(**s) ) )
	{
		r <<= 4;
		r += hexdigit( upper(**s) );
		(*s)++;
	}
	return r;
}

//=============================================================================
//	AddVar
//
//	Add a variable (ie, an equate) to the token buffer (including locals!)
//=============================================================================

OSErr AddVar(token * * t, const char * * s )
{
	char varName[VARSIZE+1];
	char * vnPtr = (char *)*s;
	unsigned long var;
	int vlen;

	if( (**s < 'A' && **s != gPrefs.local) || **s == 0x7F )
		return( ERROR_INVALVAR );

	if( FindEOW( (char **)s ) == NOERR )
	{
		vlen = *s - vnPtr;
		strncpy( varName, vnPtr, vlen );
		varName[vlen] = 0;

		if( varName[0] == gPrefs.local )	// "local" definition?
			sprintf( varName, "%s%d", varName, gGroup );

		if( gFirstPass || (vlen == 2  && varName[0] == 'P' && varName[1] == 'C') )
		{
			AddLong( t, gPC );
			return( NOERR );
		}

		if( hash_Search( gSymTblHdl, varName, &var ) )
		{
			AddLong( t, var );
			return( NOERR );
		}
	}
	return( ERROR_NOSUCHVAR );
}
